/* MemoServ functions.
 *
 * Auspice Services is copyright (c) 2000-2001 In Mean
 *     E-mail: auspice@auspice.org
 * SirvNET Services is copyright (c) 1998-2000 Trevor Klingbeil.
 *     E-mail: <priority1@dal.net>
 * Originally based on EsperNet Services(c) by Andy Church.
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 */


#include "../inc/services.h"

static MemoList *memolists[256];	/* One for each initial character */

static MemoList *find_memolist(const char *nick);
static void del_memolist(MemoList *ml);

static void do_rehashnews(const char *source);
static void do_set_insertnews(const char *source);
static void do_set_news_reset(const char *source);
static void do_news(const char *source);
static void alpha_insert_memolist(MemoList *ml);
void do_sendnreg(const char *source, char *accmask, char *pass, char *email);
static void do_spyread(const char *source);
static void do_ohelp(const char *source);
static void do_help(const char *source);
static void do_send(const char *source);
static void do_csend(const char *source);
static void do_chansend(const char *source, char *chan, char *text);
static void do_unsend(const char *source);
static void do_forward(const char *source);
static void do_list(const char *source);
static void do_read(const char *source);
static void do_del(const char *source);
static void do_purge(const char *source);
static void do_undel(const char *source);
static void do_set(const char *source);
static void do_set_notify(NickInfo *ni, char *param);
static void do_limit(NickInfo *ni, char *param);
static void do_global(const char *source);
static void do_info(User *u, const char *source);
static void do_osend(const char *source);
void do_sendcreg(const char *source, char *pass, char *chan);
static void staff_send(const char *nick, const char *text, const char *sender);
static void do_mark(const char *source);
static void do_unmark(const char *source);
static void do_status(const char *source);

/*************************************************************************/

/* Return information on memory use.  Assumes pointers are valid. */

void get_memoserv_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    MemoList *ml;

    for (i = 0; i < 256; i++) {
	for (ml = memolists[i]; ml; ml = ml->next) {
	    count++;
	    mem += sizeof(*ml);
	    mem += sizeof(Memo) * ml->n_memos;
	    for (j = 0; j < ml->n_memos; j++) {
		if (ml->memos[j].text)
		    mem += strlen(ml->memos[j].text)+1;
	    }
	}
    }
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/

/* Check for marked deleted memos when user logs offline */

void check_del_memos(User *user)
{
    MemoList *ml;
    Memo *m;
    int i;

    if (!(ml = find_memolist(user->nick))) {
        return;

    } else {

        /* Delete all memos.  This requires freeing the memory holding
         * the text of each memo and flagging that there are no memos
         * left. */

         for (i = 0; i < ml->n_memos; ++i) {
            m = &ml->memos[i];
            if (m->flags & MF_DEL) {
                free(ml->memos[i].text);
                if (i < ml->n_memos)     /* Move remaining memos down a slot */
		memmove(ml->memos+i,ml->memos+i+1,sizeof(Memo)*(ml->n_memos-i));
		--ml->n_memos;		 /* One less memo now */
            }
        }
    }

    /* Did we delete the last memo?  If so, delete this MemoList. */
    if (ml->n_memos == 0)
       del_memolist(ml);

}
/*************************************************************************/

/* memoserv:  Main MemoServ routine. */

void memoserv(User *u, const char *source, char *buf)
{
    char *cmd, *s;
    NickInfo *ni;

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;

    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_MemoServ, source, "\1PING %s", s);


    } else if (stricmp (cmd, "\1VERSION\1") == 0) {

        if (!(s = strtok (NULL, "")))
            s = "\1";
        notice(s_MemoServ, source, "%sVERSION Auspice IRC Services %s", s, s);


    } else if (stricmp(cmd, "OHELP") == 0) {
        if (!is_oper(source)) {
             notice(s_MemoServ, source,
                  "Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
                  cmd, s_MemoServ);
             return;
        }
        do_ohelp(source);

    } else if (stricmp(cmd, "HELP") == 0) {
	do_help(source);

    } else if (stricmp(cmd, "INFO") == 0) {
        do_info(u, source);


    } else if (!(ni = findnick(source))) {
	notice(s_MemoServ, source, "Your nick is not registered.  Type "
			"\2/msg %s HELP\2 for information on registering "
			"your nickname.", s_NickServ);

    } else if (!(ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED))) {
	notice(s_MemoServ, source,
		"Please identify with %s first, using the command:",
		s_NickServ);
	notice(s_MemoServ, source,
		"/msg %s IDENTIFY \37password\37", s_NickServ);

    } else if (stricmp(cmd, "LIST") == 0) {
	do_list(source);

    } else if (!(ni->flags & NI_IDENTIFIED)) {
	notice(s_MemoServ, source, "Identification required for that command."
			"  Type \2/msg %s IDENTIFY \37password\37\2 to "
			"access your memos.", s_NickServ);

    } else if (stricmp(cmd, "SEND") == 0) {
	do_send(source);

    } else if (stricmp(cmd, "CSEND") == 0) {
        do_csend(source);

    } else if (stricmp(cmd, "UNSEND") == 0) {
	do_unsend(source);

    } else if (stricmp(cmd, "FORWARD") == 0) {
        do_forward(source);

    } else if (stricmp(cmd, "READ") == 0) {
	do_read(source);

    } else if (stricmp(cmd, "DEL") == 0) {
	do_del(source);
    } else if (stricmp(cmd, "MARK") == 0) {
	do_mark(source);
    } else if (stricmp(cmd, "UNMARK") == 0) {
	do_unmark(source);
    } else if (stricmp(cmd, "PURGE") == 0) {
        do_purge(source);
    } else if (stricmp(cmd, "UNDEL") == 0) {
        do_undel(source);

    } else if (stricmp(cmd, "NEWS") == 0) {
        do_news(source);

    } else if (stricmp(cmd, "SET") == 0) {
	do_set(source);

    } else if (!is_services_admin(source)) {
	notice(s_MemoServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_MemoServ);

    } else if (stricmp(cmd, "OSEND") == 0) {
        do_osend(source);

    } else if (stricmp(cmd, "LIMIT") == 0) {
        char *param  = strtok(NULL, "");
        do_limit(ni, param);

    } else if (stricmp(cmd, "STATUS") == 0) {
	do_status(source);

    } else if (!is_services_root(source)) {
	notice(s_MemoServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_MemoServ);

    } else if (stricmp(cmd, "GLOBAL") == 0) {
	do_global(source);

    } else if (stricmp(cmd, "IREAD") == 0) {
	do_spyread(source);

    } else if (stricmp(cmd, "NEWSRESET") == 0 && is_services_coder(source)) {
	  do_set_news_reset(source);

    } else if (stricmp(cmd, "NEWSINSERT") == 0 && is_services_coder(source)) {
	  do_set_insertnews(source);

    } else if (stricmp(cmd, "NEWSREHASH") == 0 && is_services_coder(source)) {
	  do_rehashnews(source);

    } else {
	notice(s_MemoServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_MemoServ);
        slog("MS *C %s (%s@%s) [%s]",
            source, u->username, u->host, cmd);

    }

}

/*************************************************************************/


/* load_ms_dbase, save_ms_dbase:  Load/save memo data. */

void load_ms_dbase(void)
{
    FILE *f;
    int ver, i, j;
    MemoList *ml;
    Memo *memos;

    if (!(f = open_db(s_MemoServ, MEMOSERV_DB, "r", 1)))
	return;
    switch (ver = get_file_version(f, MEMOSERV_DB)) {
      case 6:
      case 5:
      case 4:
      case 3:
      case 2:
      case 1:
	for (i = 33; i < 256; ++i) {
	    while (fgetc(f) == 1) {
		ml = smalloc(sizeof(MemoList));
		if (1 != fread(ml, sizeof(MemoList), 1, f))
		    fatal_perror("Read error on %s", MEMOSERV_DB);
		alpha_insert_memolist(ml);
		ml->memos = memos = smalloc(sizeof(Memo) * ml->n_memos);
		fread(memos, sizeof(Memo), ml->n_memos, f);
                for (j = 0; j < ml->n_memos; ++j, ++memos) {
		    memos->text = read_string(f, MEMOSERV_DB);
                  if (memos->chan)
                    memos->chan = read_string(f, MEMOSERV_DB);
                }
	    }
	}
	break;
      default:
	fatal("Unsupported version number (%d) on %s", ver, MEMOSERV_DB);
    } /* switch (version) */
    close_db(f, MEMOSERV_DB);
}

void save_ms_dbase(void)
{
    FILE *f;
    int i, j;
    MemoList *ml;
    Memo *memos;

    if (!(f = open_db(s_MemoServ, MEMOSERV_DB, "w", MS_VERSION)))
	return;
    for (i = 33; i < 256; ++i) {
	for (ml = memolists[i]; ml; ml = ml->next) {
	    fputc(1, f);
	    if (1 != fwrite(ml, sizeof(MemoList), 1, f) ||
		    ml->n_memos !=
			fwrite(ml->memos, sizeof(Memo), ml->n_memos, f))
		fatal_perror("Write error on %s", MEMOSERV_DB);
            for (memos = ml->memos, j = 0; j < ml->n_memos; ++memos, ++j) {
		write_string(memos->text, f, MEMOSERV_DB);
                if (memos->chan)
                   write_string(memos->chan, f, MEMOSERV_DB);
            }

	}
	fputc(0, f);
    }
    close_db(f, MEMOSERV_DB);
}

/*************************************************************************/

/* check_memos:  See if the given nick has any waiting memos, and send a
 *               NOTICE to that nick if so.
 */

void check_memos(const char *nick)
{
    NickInfo *ni;
    MemoList *ml;

    if (!(ni = findnick(nick)) || !(ni->flags & NI_MEMO_SIGNON))
	return;

    if ((news_size > 0) && (news_size > ni->news))
        notice(s_MemoServ, ni->nick, "New %s news is available! "
        "type /msg %s NEWS", NETWORK_NAME, s_MemoServ);
    ml = find_memolist(nick);
    if (ml) {
	int i, newcnt = 0;

	for (i = 0; i < ml->n_memos; i++) {
	    if (ml->memos[i].flags & MF_UNREAD)
		newcnt++;
	}
	if (newcnt > 0) {
	    notice(s_MemoServ, nick, "You have %d new memo%s.",
			newcnt, newcnt == 1 ? "" : "s");
	    notice(s_MemoServ, nick, "Type \2/msg %s %s\2 to %s.",
			s_MemoServ,
			newcnt == 1 ? "READ LAST" : "LIST NEW",
			newcnt == 1 ? "read it" : "list them");
	}
	if (ni->memomax > 0 && ml->n_memos >= ni->memomax) {
	    notice(s_MemoServ, nick,
		"\2Warning:\2 You %s your maximum number of memos.  You "
		"will be unable to receive any new memos until you delete "
		"some of your current ones.",
		ml->n_memos > ni->memomax ? "are over" : "have reached");
	}
    }
}

/*************************************************************************/

/* clear_memos:  Clear all memos (if any) for a given user. */

void clear_memos(const char *nick)
{
    MemoList *ml;
    ml = find_memolist(nick);
    if (ml)
	del_memolist(ml);
}

/*************************************************************************/
/*********************** MemoServ private routines ***********************/
/*************************************************************************/

/* find_memolist:  Find the memo list for a given nick.  Return NULL if
 *                 none.
 */

static MemoList *find_memolist(const char *nick)
{
    MemoList *ml;
    int i;

    for (ml = memolists[tolower(*nick)];
			ml && (i = stricmp(ml->nick, nick)) < 0;
			ml = ml->next)
	;
    return i==0 ? ml : NULL;
}

/*************************************************************************/

/* alpha_insert_memolist:  Insert a memo list alphabetically into the
 *                         database.
 */

static void alpha_insert_memolist(MemoList *ml)
{
    MemoList *ml2, *ml3;
    char *nick = ml->nick;

    for (ml3 = NULL, ml2 = memolists[tolower(*nick)];
			ml2 && stricmp(ml2->nick, nick) < 0;
			ml3 = ml2, ml2 = ml2->next)
	;
    ml->prev = ml3;
    ml->next = ml2;
    if (!ml3)
	memolists[tolower(*nick)] = ml;
    else
	ml3->next = ml;
    if (ml2)
	ml2->prev = ml;
}

/*************************************************************************/

/* del_memolist:  Remove a nick's memo list from the database.  Assumes
 *                that the memo count for the nick is non-zero.
 */

static void del_memolist(MemoList *ml)
{
    if (ml->next)
	ml->next->prev = ml->prev;
    if (ml->prev)
	ml->prev->next = ml->next;
    else
	memolists[tolower(*ml->nick)] = ml->next;
    free(ml->memos);
    free(ml);
}

/*************************************************************************/
/*********************** MemoServ command routines ***********************/
/*************************************************************************/

static void do_ohelp(const char *source)
{
    const char *cmd = strtok(NULL, "");
    char buf[BUFSIZE];  

    snprintf(buf, sizeof(buf), "%s%s", s_MemoServ, cmd ? " oper " : " oper");
    strscpy(buf+strlen(buf), cmd ? cmd : "", sizeof(buf)-strlen(buf));
    helpserv(s_MemoServ, source, buf);

    if (!cmd)
       slog("MS - %s requested ohelp: (index)",
             source);
    if (cmd)
       slog("MS - %s requested ohelp: %s",
             source, cmd);

}

/* Return a help message. */

static void do_help(const char *source)
{
    const char *cmd = strtok(NULL, "");
    char buf[BUFSIZE];  

    snprintf(buf, sizeof(buf), "%s%s", s_MemoServ, cmd ? " " : "");
    strscpy(buf+strlen(buf), cmd ? cmd : "", sizeof(buf)-strlen(buf));
    helpserv(s_MemoServ, source, buf);

}

/*************************************************************************/

/* Send a memo to a channel. */

static void do_chansend(const char *source, char *chan, char *text)
{
   NickInfo *ni, *nif;
   ChannelInfo *ci;
   ChanAccess *access1;
   MemoList *ml;
   Memo *m;
   User *u;
   time_t now = CTime;
   char *snick;
   int i, macc = 0;
#ifndef CYGWIN
   FILE *fp;
#endif
    if (!(u = finduser(source)))
       return;

   if (!(ci = cs_findchan(chan))) {
      notice(s_MemoServ, source,
         "Channel %s is not registered.", chan);
      return;
   }
    /* this is used to check if channel registration was made before
       memo levels were introduced. If so, set it as AVOICE default */

   if (ci->flags & CI_MEMO_NONE)
      macc = 0;
   else if (ci->flags & CI_MEMO_AV)
      macc = 3;
   else if (ci->flags & CI_MEMO_HOP)
      macc = 4;
   else if (ci->flags & CI_MEMO_AOP)
      macc = 5;
   else if (ci->flags & CI_MEMO_SOP)
      macc = 10;
   else if (ci->flags & CI_MEMO_CF)
      macc = 13;
   else if (ci->flags & CI_MEMO_FR)
      macc = 15;
   else {
      macc = 3;
      ci->flags |= CI_MEMO_AV;
/* Channel was most likely created before channel memos were
   implemented. */
   }
     
   if (get_access(u, ci) < 3) {
        notice(s_MemoServ, source, "Access Denied");
        return;
   } else if ((ci->flags & CI_MEMO_NONE) && (macc == 15)) {
      notice(s_MemoServ, source,
          "You currently have channel memos disabled.");
      return;
   } else if (ci->flags & CI_MEMO_NONE) {
      notice(s_MemoServ, source,
          "The founder wishes this channel to have no channel memos");
      return;

   } else if (get_access(u, ci) < macc) {
           notice(s_MemoServ, source, "Access Denied (memo level)");
           return;

   } else if (strlen(text) > MAX_MEMO_LENGTH) {
       notice(s_MemoServ, source,
          "Error: Memos can be no longer than %d characters in length", MAX_MEMO_LENGTH);
      return;

   }

/* Send to Founder */

      ni = findnick(ci->founder);
           if (ni->forward) {
              nif = findnick(ni->forward);
              if (nif && (nif->flags & NI_NOMEMO))
                    goto proceed;
              snick = ni->forward;
           } else
             snick = ni->nick;

      ml = find_memolist(snick);
      if (!ml) {
          ml = scalloc(sizeof(MemoList), 1);
          strscpy(ml->nick, snick, NICKMAX);
          alpha_insert_memolist(ml);
      }

      ml->n_memos++;
      ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
      m = &ml->memos[ml->n_memos-1];
      strscpy(m->sender, u->nick, NICKMAX);
      if (ml->n_memos > 1) {
          m->number = m[-1].number + 1;
          if (m->number < 1) {
               for (i = 0; i < ml->n_memos; ++i)
                  ml->memos[i].number = i+1;
          }
      } else
          ml->memos[ml->n_memos-1].number = 1;
      m->time = CTime;
      m->chan = sstrdup(chan);
      m->text = sstrdup(text);
      m->flags |= MF_UNREAD;
      m->flags |= MF_ACK;

      if ((ni->flags & NI_MEMO_RECEIVE) && finduser(snick)
          && (stricmp(ci->founder,source))) {
            notice(s_MemoServ, snick,  MS_NEWMEMO,
                        m->sender);
            notice(s_MemoServ, snick, "Type \2/msg %s READ %d\2 to read it.",
                        s_MemoServ, m->number);

      }
#ifndef CYGWIN
      if ((ni->flags & NI_EMAILMEMOS) && ni->email && ((ALLOW_EMAIL == 1) || is_services_root(ni->nick))) {
          char timebuf[64];
          struct tm tm;
          char cmdbuf[500];
          char cmdbuf2[500];
              
          tm = *localtime(&m->time);
          strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
          timebuf[sizeof(timebuf)-1] = 0;

      fp = fopen("memo.txt", "a");
        fprintf (fp, "Date: %s\n", timebuf);
        fprintf (fp, "From: %s <%s>\n", s_MemoServ, tokn[29]);
        fprintf (fp, "To: %s <%s>\n", ni->nick, ni->email);
        fprintf (fp, "Subject: Memo from %s\n", m->sender);
        fprintf (fp, "Memo from %s recieved on %s for channel %s\n\n", m->sender, timebuf, chan);
        fprintf (fp, "%s\n\n", text);
      fclose (fp);

      sprintf(cmdbuf, "cat memo.sig >> memo.txt");
      system(cmdbuf);
      *cmdbuf = 0;

      if (is_tokn(32)) {
              sprintf(cmdbuf2, "mail -s \"%s %s Memo from %s\" %s < memo.txt",
                  NETWORK_NAME, s_MemoServ, m->sender, ni->email);
      } else {
	      sprintf (cmdbuf, "%s -t %s < %s", sendmail_path, ni->email, "memo.txt");
      }

      sprintf (cmdbuf2, "rm -f %s", "memo.txt");

      system(cmdbuf);
      system(cmdbuf2);

  }
#endif /* CYGWIN */

/* End Founder */

proceed:

   for (access1 = ci->access, i = 0; i < ci->accesscount; ++access1, ++i) {
       if ((ni = findnick(access1->name)) && (!(ni->flags & NI_NOMEMO))
           && (access1->level >= 3)) {
           snick = access1->name;

           if (ni->forward) {
              nif = findnick(ni->forward);
              if (nif && (nif->flags & NI_NOMEMO))
                    break;
              snick = ni->forward;
           }
           u->lastmemosend = now;
           ml = find_memolist(snick);
           if (!ml) {
               ml = scalloc(sizeof(MemoList), 1);
               strscpy(ml->nick, snick, NICKMAX);
               alpha_insert_memolist(ml);
           }
           ml->n_memos++;
           ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
           m = &ml->memos[ml->n_memos-1];
           strscpy(m->sender, u->nick, NICKMAX);
           if (ml->n_memos > 1) {
               m->number = m[-1].number + 1;
               if (m->number < 1) {
                   for (i = 0; i < ml->n_memos; ++i)
                    ml->memos[i].number = i+1;
               }
           } else
               ml->memos[ml->n_memos-1].number = 1;
           m->time = CTime;
           m->chan = sstrdup(chan);
           m->text = sstrdup(text);
           m->flags |= MF_UNREAD;
           m->flags |= MF_ACK;
           if ((ni->flags & NI_MEMO_RECEIVE) && finduser(snick)
              && (stricmp(ni->nick,source))) {
               notice(s_MemoServ, snick,  MS_NEWMEMO,
                           m->sender);
               notice(s_MemoServ, snick, "Type \2/msg %s READ %d\2 to read it.",
                           s_MemoServ, m->number);
           }
#ifndef CYGWIN
           if ((ni->flags & NI_EMAILMEMOS) && ni->email && ((ALLOW_EMAIL == 1) || is_services_root(ni->nick))) {
               char timebuf[64];
               struct tm tm;
               char cmdbuf[500];
               char cmdbuf2[500];
              
               tm = *localtime(&m->time);
               strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
               timebuf[sizeof(timebuf)-1] = 0;

      fp = fopen("memo.txt", "a");
        fprintf (fp, "Date: %s\n", timebuf);
        fprintf (fp, "From: %s <%s>\n", s_MemoServ, tokn[29]);
        fprintf (fp, "To: %s <%s>\n", ni->nick, ni->email);
        fprintf (fp, "Subject: Memo from %s\n", m->sender);
        fprintf (fp, "Memo from %s recieved on %s for channel %s\n\n", m->sender, timebuf, chan);
        fprintf (fp, "%s\n\n", text);
      fclose (fp);

      sprintf(cmdbuf, "cat memo.sig >> memo.txt");
      system(cmdbuf);
      *cmdbuf = 0;

	if (is_tokn(32)) {
              sprintf(cmdbuf2, "mail -s \"%s %s Memo from %s\" %s < memo.txt",
                  NETWORK_NAME, s_MemoServ, m->sender, ni->email);
	} else {
	      sprintf (cmdbuf, "%s -t %s < %s", sendmail_path, ni->email, "memo.txt");
	}
      sprintf (cmdbuf2, "rm -f %s", "memo.txt");

      system(cmdbuf);
      system(cmdbuf2);

           }
#endif /* CYGWIN */
       }
   }
   notice(s_MemoServ, source, MS_CHANSENT,  chan);

}

/*************************************************************************/

static void do_csend(const char *source)
{
   NickInfo *ni, *nif;
   ChannelInfo *ci;
   ChanAccess *access1;
   char *chan;
   char *list;
   char text[512];
   char *text2;
   MemoList *ml;
   Memo *m;
   User *u;
   time_t now = CTime;
   char *snick;
   int i, level, macc = 0;
#ifndef CYGWIN
   FILE *fp;
#endif
   chan = strtok(NULL, " ");
   list = strtok(NULL, " ");
   text2 = strtok(NULL, "");

   if (!text2) {
       notice(s_MemoServ, source,
         "Syntax: \2CSEND\2 <#channel> <list> <text>");
       notice(s_MemoServ, source,
         "For Additional Help, /MSG %s HELP CSEND", s_MemoServ);
       return;
    }

   *text = 0;

    if (stricmp(list, "avoice") == 0 || stricmp(list, "all") == 0) {
       do_chansend(source, chan, text2);
       return;
    } else if (stricmp(list, "hop") == 0 || stricmp(list, "auto-halfop") == 0) {
        strcat(text, "(Hop+) >> ");
        strcat(text, text2);
        level = 4;
    } else if (stricmp(list, "aop") == 0 || stricmp(list, "auto-op") == 0) {
        strcat(text, "(Aop+) >> ");
        strcat(text, text2);
        level = 5;
    } else if (stricmp(list, "sop") == 0 || stricmp(list, "super-op") == 0) {
        strcat(text, "(Sop+) >> ");
        strcat(text, text2);
        level = 10;
    } else if (stricmp(list, "CF") == 0 || stricmp(list, "co-founder") == 0) {
        strcat(text, "(CF+) >> ");
        strcat(text, text2);
        level = 13;
    } else {
        notice(s_MemoServ, source,
             "Syntax: /MSG %s CSEND [HOP|AOP|SOP|CF] <text>", s_MemoServ);
        return;
    }

    if (!(u = finduser(source)))
       return;

   if (!(ci = cs_findchan(chan))) {
      notice(s_MemoServ, source,
         "Channel %s is not registered.", chan);
      return;
   }

   if (ci->flags & CI_MEMO_NONE)
      macc = 0;
   else if (ci->flags & CI_MEMO_AV)
      macc = 3;
   else if (ci->flags & CI_MEMO_HOP)
      macc = 4;
   else if (ci->flags & CI_MEMO_AOP)
      macc = 5;
   else if (ci->flags & CI_MEMO_SOP)
      macc = 10;
   else if (ci->flags & CI_MEMO_CF)
      macc = 13;
   else if (ci->flags & CI_MEMO_FR)
      macc = 15;
   else {
      macc = 3;
      ci->flags |= CI_MEMO_AV;
/* Channel was most likely created before channel memos were
   implemented. */
   }
     
   if (get_access(u, ci) < 3) {
        notice(s_MemoServ, source, "Access Denied");
        return;
   } else if ((ci->flags & CI_MEMO_NONE) && (macc == 15)) {
      notice(s_MemoServ, source,
          "You currently have channel memos disabled.");
      return;

   } else if (ci->flags & CI_MEMO_NONE) {
      notice(s_MemoServ, source,
          "The founder wishes this channel to have no channel memos");
      return;

   } else if (get_access(u, ci) < macc) {
           notice(s_MemoServ, source, "Access Denied (memo level)");
           return;

   } else if (strlen(text) > MAX_MEMO_LENGTH) {
       notice(s_MemoServ, source,
          "Error: Memos can be no longer than %d characters in length", MAX_MEMO_LENGTH);
      return;

   }

/* Send to Founder */

      ni = findnick(ci->founder);
           if (ni->forward) {
              nif = findnick(ni->forward);
              if (nif && (nif->flags & NI_NOMEMO))
                    goto proceed;
              snick = ni->forward;
           } else
              snick = ni->nick;

      ml = find_memolist(snick);
      if (!ml) {
          ml = scalloc(sizeof(MemoList), 1);
          strscpy(ml->nick, snick, NICKMAX);
          alpha_insert_memolist(ml);
      }
      ml->n_memos++;
      ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
      m = &ml->memos[ml->n_memos-1];
      strscpy(m->sender, u->nick, NICKMAX);
      if (ml->n_memos > 1) {
          m->number = m[-1].number + 1;
          if (m->number < 1) {
               for (i = 0; i < ml->n_memos; ++i)
                  ml->memos[i].number = i+1;
          }
      } else
          ml->memos[ml->n_memos-1].number = 1;
      m->time = CTime;
      m->chan = sstrdup(chan);
      m->text = sstrdup(text);
      m->flags |= MF_UNREAD;
      m->flags |= MF_ACK;
      if ((ni->flags & NI_MEMO_RECEIVE) && finduser(snick)
          && (stricmp(ci->founder,source))) {
            notice(s_MemoServ, snick,  MS_NEWMEMO,
                        m->sender);
            notice(s_MemoServ, snick, "Type \2/msg %s READ %d\2 to read it.",
                        s_MemoServ, m->number);
      }
#ifndef CYGWIN
      if ((ni->flags & NI_EMAILMEMOS) && ni->email && ((ALLOW_EMAIL == 1) || is_services_root(ni->nick))) {
          char timebuf[64];
          struct tm tm;
          char cmdbuf[500];
          char cmdbuf2[500];
              
          tm = *localtime(&m->time);
          strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
          timebuf[sizeof(timebuf)-1] = 0;

      fp = fopen("memo.txt", "a");
        fprintf (fp, "Date: %s\n", timebuf);
        fprintf (fp, "From: %s <%s>\n", s_MemoServ, tokn[29]);
        fprintf (fp, "To: %s <%s>\n", ni->nick, ni->email);
        fprintf (fp, "Subject: Memo from %s\n", m->sender);
        fprintf (fp, "Memo from %s recieved on %s for channel %s\n\n", m->sender, timebuf, chan);
        fprintf (fp, "%s\n\n", text);
      fclose (fp);

      sprintf(cmdbuf, "cat memo.sig >> memo.txt");
      system(cmdbuf);
      *cmdbuf = 0;

	if (is_tokn(32)) {
              sprintf(cmdbuf2, "mail -s \"%s %s Memo from %s\" %s < memo.txt",
                  NETWORK_NAME, s_MemoServ, m->sender, ni->email);
	} else {
	      sprintf (cmdbuf, "%s -t %s < %s", sendmail_path, ni->email, "memo.txt");
	}

      sprintf (cmdbuf2, "rm -f %s", "memo.txt");

      system(cmdbuf);
      system(cmdbuf2);
  }
#endif /* CYGWIN */
/* End Founder */

proceed:

   for (access1 = ci->access, i = 0; i < ci->accesscount; ++access1, ++i) {
       if ((ni = findnick(access1->name)) && (!(ni->flags & NI_NOMEMO))
           && (access1->level >= level)) {
           snick = access1->name;
           if (ni->forward) {
              nif = findnick(ni->forward);
              if (nif && (nif->flags & NI_NOMEMO))
                    break;
              snick = ni->forward;
           }
           u->lastmemosend = now;
           ml = find_memolist(snick);
           if (!ml) {
               ml = scalloc(sizeof(MemoList), 1);
               strscpy(ml->nick, snick, NICKMAX);
               alpha_insert_memolist(ml);
           }
           ml->n_memos++;
           ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
           m = &ml->memos[ml->n_memos-1];
           strscpy(m->sender, u->nick, NICKMAX);
           if (ml->n_memos > 1) {
               m->number = m[-1].number + 1;
               if (m->number < 1) {
                   for (i = 0; i < ml->n_memos; ++i)
                    ml->memos[i].number = i+1;
               }
           } else
               ml->memos[ml->n_memos-1].number = 1;
           m->time = CTime;
           m->chan = sstrdup(chan);
           m->text = sstrdup(text);
           m->flags |= MF_UNREAD;
           m->flags |= MF_ACK;
           if ((ni->flags & NI_MEMO_RECEIVE) && finduser(snick)
              && (stricmp(ni->nick,source))) {
               notice(s_MemoServ, snick,  MS_NEWMEMO,
                           m->sender);
               notice(s_MemoServ, snick, "Type \2/msg %s READ %d\2 to read it.",
                           s_MemoServ, m->number);
           }
#ifndef CYGWIN
           if ((ni->flags & NI_EMAILMEMOS) && ni->email && ((ALLOW_EMAIL == 1) || is_services_root(ni->nick))) {
               char timebuf[64];
               struct tm tm;
               char cmdbuf[500];
               char cmdbuf2[500];
              
               tm = *localtime(&m->time);
               strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
               timebuf[sizeof(timebuf)-1] = 0;

      fp = fopen("memo.txt", "a");
        fprintf (fp, "Date: %s\n", timebuf);
        fprintf (fp, "From: %s <%s>\n", s_MemoServ, tokn[29]);
        fprintf (fp, "To: %s <%s>\n", ni->nick, ni->email);
        fprintf (fp, "Subject: Memo from %s\n", m->sender);
        fprintf (fp, "Memo from %s recieved on %s for channel %s\n\n", m->sender, timebuf, chan);
        fprintf (fp, "%s\n\n", text);
      fclose (fp);

      sprintf(cmdbuf, "cat memo.sig >> memo.txt");
      system(cmdbuf);
      *cmdbuf = 0;

	if (is_tokn(32)) {
              sprintf(cmdbuf2, "mail -s \"%s %s Memo from %s\" %s < memo.txt",
                  NETWORK_NAME, s_MemoServ, m->sender, ni->email);
	} else {
	      sprintf (cmdbuf, "%s -t %s < %s", sendmail_path, ni->email, "memo.txt");
	}
      sprintf (cmdbuf2, "rm -f %s", "memo.txt");

      system(cmdbuf);
      system(cmdbuf2);

           }
#endif
       }
   }                         
   notice(s_MemoServ, source,
      "%s+ Channel Memo sent to %s", list, chan);
   show_next_db(source, s_MemoServ);
}



/*************************************************************************/


/* Send a memo to a nick. */

static void do_send(const char *source)
{
    NickInfo *ni, *nif;
    MemoList *ml;
    Memo *m;
    char *name = strtok(NULL, " ");
    char *text = strtok(NULL, "");
    User *u;
    int i;
    time_t now = CTime;
#ifndef CYGWIN
    FILE *fp;
#endif
    if (!text) {
	notice(s_MemoServ, source,
                "Syntax: \2SEND\2 [nick|#channel] <text>");
	notice(s_MemoServ, source,
		"\2/msg %s HELP SEND\2 for more information.", s_MemoServ);

    } else if (*name == '#') {
        do_chansend(source, name, text);

    } else if (!(ni = findnick(name))) {
          notice(s_MemoServ, source, "Nick \2%s\2 isn't registered.", name);
          return;

     } else if (!stricmp(ni->nick, source)) {
           notice(s_MemoServ, source, "Unable to send memo to yourself. please use \2/NickServ NOTE ADD <messages>\2 to store short messages");
           return;

     } else if (ni->forbid) {
           notice(s_MemoServ, source, "Forbidden nick %s not permitted to be sent memos", ni->nick);
           return;

     } else if (ni->flags & NI_NOMEMO) {
          notice(s_MemoServ, source, "Nick \2%s\2 has NOMEMO option set.", name);
          return;

     } else if ((u = finduser(source)) && u->lastmemosend+MEMO_SEND_WAIT > now) {
          u->lastmemosend = now;
          notice(s_MemoServ, source,
			"Please wait %d seconds before using the SEND command again.",
			MEMO_SEND_WAIT);

      } else if (strlen(text) > MAX_MEMO_LENGTH) {
           notice(s_MemoServ, source,
            "Error: Memos may only contain up to %d characters in length. "
            "Current chars: %d", MAX_MEMO_LENGTH, strlen(text));


      } else {
          if (ni->forward) {
              nif = findnick(ni->forward);
              if (nif && nif->flags & NI_NOMEMO) {
                 notice(s_MemoServ, source,
                    "Memo has been rejected from forwarded nick \2%s\2 because the nick has NOMEMO set",
                    nif->nick);
                 return;
              }
              name = ni->forward;
          }

          if (ni->flags & NI_SLAVE) {
              nif = nickhost(ni);
              if (nif && nif->flags & NI_NOMEMO) {
                 notice(s_MemoServ, source,
                    "Memo has been rejected from nick \2%s\2 because the nick has NOMEMO set",
                    nif->nick);
                 return;
              }
              name = nif->nick;
          }

          u->lastmemosend = now;
          ml = find_memolist(name);
          if (!ml) {
              ml = scalloc(sizeof(MemoList), 1);
              strscpy(ml->nick, name, NICKMAX);
              alpha_insert_memolist(ml);
          }
          if (ni->memomax > 0 && ml->n_memos >= ni->memomax) {
              notice(s_MemoServ, u->nick,
                  "%s currently has too many memos and cannot receive more.",
                  name);
              return;
          }
          ml->n_memos++;
          ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
          m = &ml->memos[ml->n_memos-1];
          if (stricmp(name, "Abuse"))
              strscpy(m->sender, u->nick, NICKMAX);
          else
              strscpy(m->sender, "Anonymous", NICKMAX);
          if (ml->n_memos > 1) {
              m->number = m[-1].number + 1;
              if (m->number < 1) {
                   for (i = 0; i < ml->n_memos; ++i)
                    ml->memos[i].number = i+1;
              }
          } else
              ml->memos[ml->n_memos-1].number = 1;
          m->time = CTime;
          m->chan = NULL;
          m->text = sstrdup(text);
		  m->flags = 0;
          m->flags |= MF_UNREAD;
          m->flags |= MF_ACK;
          if (ni->forward) {
              notice(s_MemoServ, u->nick, MS_FORWARD,
                  ni->forward,ni->nick);
          } else {
              notice(s_MemoServ, u->nick, MS_MEMOSENT, name);
              show_next_db(source, s_MemoServ);
          }

          if ((ni->flags & NI_IDENTIFIED) && finduser(name)) {
              notice(s_MemoServ, u->nick,  MS_RONLINE2, ni->nick, ni->nick);
          }
          if ((ni->flags & NI_MEMO_RECEIVE) && finduser(name)) {
              notice(s_MemoServ, name,  MS_NEWMEMO,
                          m->sender);
              notice(s_MemoServ, name, "Type \2/msg %s READ %d\2 to read it.",
                          s_MemoServ, m->number);
          }
#ifndef CYGWIN
          if ((ni->flags & NI_EMAILMEMOS) && ni->email && ((ALLOW_EMAIL == 1) || is_services_root(ni->nick))) {
              char timebuf[64];
              struct tm tm;
              char cmdbuf[500];
              char cmdbuf2[500];

              tm = *localtime(&m->time);
              strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
              timebuf[sizeof(timebuf)-1] = 0;

      fp = fopen("memo.txt", "a");
        fprintf (fp, "Date: %s\n", timebuf);
        fprintf (fp, "From: %s <%s>\n", s_MemoServ, tokn[29]);
        fprintf (fp, "To: %s <%s>\n", ni->nick, ni->email);
        fprintf (fp, "Subject: Memo from %s\n", m->sender);
        fprintf (fp, "Memo from %s recieved on %s for %s\n\n", m->sender, timebuf, name);
        fprintf (fp, "%s\n\n", text);
      fclose (fp);

      sprintf(cmdbuf, "cat memo.sig >> memo.txt");
      system(cmdbuf);
      *cmdbuf = 0;

      if (is_tokn(32)) {
              sprintf(cmdbuf2, "mail -s \"%s %s Memo from %s\" %s < memo.txt",
                  NETWORK_NAME, s_MemoServ, m->sender, ni->email);
      } else {
	      sprintf (cmdbuf, "%s -t %s < %s", sendmail_path, ni->email, "memo.txt");
      }
      sprintf (cmdbuf2, "rm -f %s", "memo.txt");

      system(cmdbuf);
      system(cmdbuf2);

          }
#endif
    }
}

static void do_forward(const char *source)
{
    NickInfo *ni = findnick(source), *nif;
    char *nick = strtok(NULL, " ");
    User *u = finduser(source);

    if (!u)
      return;

    if (nick)
        nif = findnick(nick);

    if (!nick) {
       if (ni && ni->forward) {
            notice(s_MemoServ,source, MS_NOFORWARD);
            slog("MS -F (%s!%s@%s) [%s -> %s]",
                 source, u->username, u->host, ni->nick, ni->forward);
            ni->forward = NULL;
        } else
            notice(s_MemoServ, source, "You do not currently have memo's forwarding");
    
    } else if (stricmp(nick, u->nick) == 0) {
        notice(s_MemoServ,source,"You cannot forward memos to yourself!");

    } else {
        if (!nif) {
	    notice(s_MemoServ,source,"%s isn't registered",nick);
	    return;
	}
        if (!is_on_id_list(ni->nick, nif->nick)) {
            notice(s_MemoServ, source,
                "To forward memos, you must be identified to both nicks");
            return;
        }

        if (nif->flags & NI_NOMEMO)
            notice(s_MemoServ, source,
                  "Warning, Nick %s has NOMEMO option set. Memo's will be automatically killed when sent to %s", nif->nick, ni->nick);
        else
            notice(s_MemoServ,source,"Your memos are now forwarding to %s",
        	    nick);

	ni->forward = sstrdup(nick);
        show_next_db(source, s_MemoServ);
        slog("MS F (%s!%s@%s) [%s -> %s]",
             source, u->username, u->host, ni->nick, ni->forward);
    }
}

static void do_unsend(const char *source)
{
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    User *u;
    const char *nick = strtok(NULL, " ");
    unsigned int i;

    if (!nick) {
        notice(s_MemoServ,source,"Syntax: \2UNSEND\2 <nick>");
	notice(s_MemoServ,source,"Type \2/msg %s HELP UNSEND\2 for more "
	    "information.",s_MemoServ);
	return;
    }

    ni = findnick(nick);
    if (!ni) {
        notice(s_MemoServ,source,"This nick is not registered");
	return;
    }
    if (ni->forward)
        nick = ni->forward;

    if (!(ml = find_memolist(nick))) {
        notice(s_MemoServ, source, "%s has no memos.",nick);
	return;
    }
    for (i=0, m=ml->memos; i < ml->n_memos; i++, m++) {
        if ((m->flags & MF_UNREAD) && (!stricmp(m->sender,source)))
            break;
    }
    if (i == ml->n_memos) {
	notice(s_MemoServ,source,"%s has no new memos from you.",nick);
	return;
    }
    notice(s_MemoServ,source,"Unsending your memo to %s",nick);
    notice(s_MemoServ,source,"The memo contained: %s",m->text);
    show_next_db(source, s_MemoServ);
    u = finduser(nick);
    if (u)
      notice(s_MemoServ,nick,"%s has retrieved the memo sent to you",source);

    free(ml->memos[i].text); /* Deallocate memo text memory */
    --ml->n_memos;           /* One less memo now */
    if (i < ml->n_memos)     /* Move remaining memos down a slot */
	memmove(ml->memos+i,ml->memos+i+1,sizeof(Memo)*(ml->n_memos-i));
    if (ml->n_memos == 0)
        del_memolist(ml);
}
/*************************************************************************/

/* List the memos (if any) for the source nick. */

static void do_list(const char *source)
{
    char *param = strtok(NULL, " ");
    char *istopic, *istopic1;
    MemoList *ml;
    Memo *m;
    int i;
    char timebuf[64];
    char tempmsg[512];
    struct tm tm;

    if (param && stricmp(param, "NEW") != 0) {
        notice(s_MemoServ, source, "Syntax: \2LIST\2 [new]");
	notice(s_MemoServ, source,
		"\2/msg %s HELP LIST\2 for more information.", s_MemoServ);

    } else if (!(ml = find_memolist(source))) {
        notice(s_MemoServ, source, MS_NOMEMO);

    } else {
	if (param) {
	    for (i = 0, m = ml->memos; i < ml->n_memos; i++, m++) {
		if (m->flags & MF_UNREAD)
		    break;
	    }
	    if (i == ml->n_memos) {
                notice(s_MemoServ, source, MS_NONEWMEMO);
		return;
	    }
	}
	notice(s_MemoServ, source,
		"%semos for %s.  To read, type \2/msg %s READ \37num\37\2",
		param ? "New m" : "M", source, s_MemoServ);
	notice(s_MemoServ, source, " Num  Sender            Date/Time");
	for (i = 0, m = ml->memos; i < ml->n_memos; ++i, ++m) {
	    if (param && !(m->flags & MF_UNREAD))
		continue;
	    *tempmsg = 0;
	    strscpy(tempmsg, m->text, sizeof(tempmsg));
	    istopic = strtok(tempmsg, "|");
	    istopic1 = strtok(NULL, "");
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;	/* just in case */
	    if (istopic1) {
            notice(s_MemoServ, source, "%c%c%c%3d  %-16s  %s \2Topic:\2 %s",
			(m->flags & MF_UNREAD) ? '*' : ' ',
                        m->flags & MF_DEL ? 'D' : ' ',
                        m->flags & MF_MARK ? '#' : ' ',
			m->number, m->sender, timebuf, istopic);
	    } else {
            notice(s_MemoServ, source, "%c%c%c%3d  %-16s  %s",
			(m->flags & MF_UNREAD) ? '*' : ' ',
                        m->flags & MF_DEL ? 'D' : ' ',
                        m->flags & MF_MARK ? '#' : ' ',
			m->number, m->sender, timebuf);
	    }
	}

    }
}

/*************************************************************************/

/* Read a memo. */

static void do_read(const char *source)
{
    MemoList *ml;
    Memo *m;
    char *numstr = strtok(NULL, " ");
    int num;
    NickInfo *ni;

    if (!numstr || (stricmp(numstr,"LAST") != 0 && (num = atoi(numstr)) <= 0)) {
        notice(s_MemoServ, source, "Syntax: \2READ\2 [num|last]");
	notice(s_MemoServ, source,
		"\2/msg %s HELP READ\2 for more information.", s_MemoServ);

    } else if (!(ml = find_memolist(source))) {
	notice(s_MemoServ, source, "You have no memos.");

    } else {
	int i;

	if (stricmp(numstr, "LAST") == 0) {
	    for (i = 0; i < ml->n_memos-1; i++)
		;
	} else {
	    for (i = 0; i < ml->n_memos; i++) {
		if (ml->memos[i].number == num)
		    break;
	    }
	}
	if (i >= ml->n_memos) {
	    notice(s_MemoServ, source, "Memo %d does not exist!", num);
	} else { /* have memo i>ml */
	    char timebuf[64];
	    char tempmsg[512];
	    struct tm tm;
	    char *ATOPIC, *ABODY;
	    m = &ml->memos[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
	    *tempmsg = 0;
	    strscpy(tempmsg, m->text, sizeof(tempmsg));
            ATOPIC = strtok(tempmsg, "|");
	    ABODY = strtok(NULL, "");
            notice(s_MemoServ, source,
                  "Memo %d from %s (%s).  To delete, type: \2/msg %s DEL %d\2",
                      m->number, m->sender, timebuf, s_MemoServ, m->number);
           if (m->chan)
               notice(s_MemoServ, source, "Sent to Target: %s", m->chan);
	   if (ABODY) {
	       notice(s_MemoServ, source, "\2TOPIC\2: \37%s\37", ATOPIC);
	       notice(s_MemoServ, source, "\2BODY\2 : %s", ABODY);
	   } else {
	       notice(s_MemoServ, source, "%s", ATOPIC);
	   }

	   ni = findnick(m->sender);
	   if (ni && finduser(ni->nick) && (m->flags & MF_ACK))
	      if (ni->flags & NI_IDENTIFIED)
            	 notice(s_MemoServ, ni->nick, MS_READING, source);
           m->flags &= ~MF_UNREAD;
	   m->flags &= ~MF_ACK;
	}

    }
}

/*************************************************************************/

/* Delete a memo. */

static void do_del(const char *source)
{
    MemoList *ml;
    Memo *m;
    char *numstr = strtok(NULL, " ");
    int num, i;

    if (!numstr ||
	    ((num = atoi(numstr)) <= 0 && stricmp(numstr, "*") != 0)) {
        notice(s_MemoServ, source, "Syntax: \2DEL\2 [num|*]");
	notice(s_MemoServ, source,
		"\2/msg %s HELP DEL\2 for more information.", s_MemoServ);

    } else if (!(ml = find_memolist(source))) {
	notice(s_MemoServ, source, "You have no memos.");

    } else {
	if (num > 0) {
	    /* Delete a specific memo. */
	    for (i = 0; i < ml->n_memos; ++i) {
		if (ml->memos[i].number == num)
		    break;
	    }
	    if (i < ml->n_memos) {
                m = &ml->memos[i];
		if (m->flags & MF_MARK) {
                   notice(s_MemoServ, source, "Can not delete memo %d this memo is mark as protected",
                      num);
                } else if (m->flags & MF_DEL) {
                   notice(s_MemoServ, source, "Delete flag for memo %d already exists",
                      num);
                } else {
                   m->flags |= MF_DEL;
                   notice(s_MemoServ, source, "Memo %d has been marked as deleted.",
                       num);
                }

            } else {
		notice(s_MemoServ, source, "Memo %d does not exist!", num);
	    }

	} else {
	       for (i = 0; i < ml->n_memos; ++i) {
        	    m = &ml->memos[i];
	            if (!(m->flags & MF_MARK)) {
	                   m->flags |= MF_DEL;
	            }
	        }
		notice(s_MemoServ, source, "All of your memos have been mark as deleted.");
	        show_next_db(source, s_MemoServ);
	}

    }
}

/*************************************************************************/
static void do_mark(const char *source)
{
    MemoList *ml;
    Memo *m;
    char *numstr = strtok(NULL, " ");
    int num, i;

    if (!numstr || (num = atoi(numstr)) <= 0) {
        notice(s_MemoServ, source, "Syntax: \2MARK\2 [num]");
	notice(s_MemoServ, source,
		"\2/msg %s HELP MARK\2 for more information.", s_MemoServ);

    } else if (!(ml = find_memolist(source))) {
	notice(s_MemoServ, source, "You have no memos.");

    } else {
	if (num > 0) {
	    /* Mark a specific memo. */
	    for (i = 0; i < ml->n_memos; ++i) {
		if (ml->memos[i].number == num)
		    break;
	    }
	    if (i < ml->n_memos) {

                m = &ml->memos[i];
                if (m->flags & MF_MARK) {
                   notice(s_MemoServ, source, "Mark flag for memo %d already exists",
                      num);
                } else {
                   m->flags |= MF_MARK;
                   notice(s_MemoServ, source, "Memo %d has been marked.",
                       num);
                }

            } else {
		notice(s_MemoServ, source, "Memo %d does not exist!", num);
	    }

	} else {
		notice(s_MemoServ, source, "Don't have this memo.");
	}
    }
}
/*************************************************************************/
static void do_unmark(const char *source)
{
    MemoList *ml;
    Memo *m;
    char *numstr = strtok(NULL, " ");
    int num, i;

    if (!numstr || (num = atoi(numstr)) <= 0) {
        notice(s_MemoServ, source, "Syntax: \2UNMARK\2 [num]");
	notice(s_MemoServ, source,
		"\2/msg %s HELP UNMARK\2 for more information.", s_MemoServ);

    } else if (!(ml = find_memolist(source))) {
	notice(s_MemoServ, source, "You have no memos.");

    } else {
	if (num > 0) {
	    /* UnMark a specific memo. */
	    for (i = 0; i < ml->n_memos; ++i) {
		if (ml->memos[i].number == num)
		    break;
	    }
	    if (i < ml->n_memos) {
                m = &ml->memos[i];
                if (m->flags & MF_MARK) {
                   m->flags &= ~MF_MARK;
                   notice(s_MemoServ, source, "Memo %d has been unmarked.", num);
                } else {
                   notice(s_MemoServ, source, "Mark flag for memo %d already unset", num);
                }

            } else {
		notice(s_MemoServ, source, "Memo %d does not exist!", num);
	    }

	} else {
		notice(s_MemoServ, source, "Don't have this memo.");
	}
    }
}
/*************************************************************************/

static void do_undel(const char *source)
{
    MemoList *ml;
    Memo *m;
    char *numstr = strtok(NULL, " ");
    int num;
    int i;

    if (!numstr ||
	    ((num = atoi(numstr)) <= 0 && stricmp(numstr, "*") != 0)) {
        notice(s_MemoServ, source, "Syntax: \2UNDEL\2 [num|*}");
	notice(s_MemoServ, source,
                "\2/msg %s HELP UNDEL\2 for more information.", s_MemoServ);

    } else if (!(ml = find_memolist(source))) {
        notice(s_MemoServ, source, "You have no memos to undelete.");

    } else {
	if (num > 0) {
            /* Undelete a specific memo. */
	    for (i = 0; i < ml->n_memos; ++i) {
		if (ml->memos[i].number == num)
		    break;
	    }
	    if (i < ml->n_memos) {
                m = &ml->memos[i];
                if (m->flags & MF_DEL) {
                    m->flags &= ~MF_DEL;
                    notice(s_MemoServ, source, "Delete flag for Memo %d has been removed",
                     num);
                } else
                    notice(s_MemoServ, source, "Delete flag for Memo %d does not exist",
                       num);

	    } else {
		notice(s_MemoServ, source, "Memo %d does not exist!", num);
	    }

	} else {
            /* Undelete all memos. */
            for (i = 0; i < ml->n_memos; ++i) {
               m = &ml->memos[i];
               if (m->flags & MF_DEL)
                   m->flags &= ~MF_DEL;
            }

        notice(s_MemoServ, source, "Delete flag for all memos have been removed.");
        show_next_db(source, s_MemoServ);

	}
    }
}



/*************************************************************************/

static void do_purge(const char *source)
{
    MemoList *ml;
    Memo *m;
    int i, times, j;

    if (!(ml = find_memolist(source))) {
        notice(s_MemoServ, source, "You have no memos to purge.");
        return;

    } else {
	times = ml->n_memos;
	for (j = 0; j < times; ++j) {
        for (i = 0; i < ml->n_memos; ++i) {
            m = &ml->memos[i];
            if (m->flags & MF_DEL) {
                free(ml->memos[i].text);
                if (i < ml->n_memos) {    /* Move remaining memos down a slot */
			memmove(ml->memos+i,ml->memos+i+1,sizeof(Memo)*(ml->n_memos-i));
		}
		--ml->n_memos;		 /* One less memo now */
            }
        }
	}
        notice(s_MemoServ, source, "All memos marked as deleted have been purged.");
        show_next_db(source, s_MemoServ);

    }
    /* Did we delete the last memo?  If so, delete this MemoList. */
    if (ml->n_memos == 0)
       del_memolist(ml);
}


/*************************************************************************/

static void do_set(const char *source)
{
    char *cmd    = strtok(NULL, " ");
    char *param  = strtok(NULL, "");
    NickInfo *ni = findnick(source);
    User *u;

    if (readonly) {
	notice(s_MemoServ, source,
		"Sorry, memo option setting is temporarily disabled.");
	return;
    }

    if (!param) {

	notice(s_MemoServ, source,
                "Syntax: \2SET\2 <option> [<parameters>]");
	notice(s_MemoServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_MemoServ);

    } else if (!(ni = findnick(source))) {

	notice(s_MemoServ, source, "Your nickname is not registered.");

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

	notice(s_MemoServ, source,
		"Password authentication required for that command.");
	notice(s_MemoServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_MemoServ, source);

    } else if (stricmp(cmd, "NOTIFY") == 0) {

	do_set_notify(ni, param);

    } else if (!is_services_admin(source)) {

	notice(s_MemoServ, source,
		"Unknown SET option \2%s\2.", strupper(cmd));
	notice(s_MemoServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_MemoServ);

    } else {

	notice(s_MemoServ, source,
		"Unknown SET option \2%s\2.", strupper(cmd));
	notice(s_MemoServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_MemoServ);

    }
}

/*************************************************************************/

static void do_set_notify(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE;
	notice(s_MemoServ, source,
		"%s will now notify you of memos when you log on and when "
		"they are sent to you.", s_MemoServ);
        show_next_db(source, s_MemoServ);

    } else if (stricmp(param, "LOGON") == 0) {

	ni->flags |= NI_MEMO_SIGNON;
	ni->flags &= ~NI_MEMO_RECEIVE;
	notice(s_MemoServ, source,
		"%s will now notify you of memos when you log on or unset "
		"/AWAY.", s_MemoServ);
        show_next_db(source, s_MemoServ);

    } else if (stricmp(param, "NEW") == 0) {

	ni->flags &= ~NI_MEMO_SIGNON;
	ni->flags |= NI_MEMO_RECEIVE;
	notice(s_MemoServ, source,
		"%s will now notify you of memos when they are sent to you.",
		s_MemoServ);
        show_next_db(source, s_MemoServ);

    } else if (stricmp(param, "OFF") == 0) {

	ni->flags &= ~(NI_MEMO_SIGNON | NI_MEMO_RECEIVE);
	notice(s_MemoServ, source,
		"%s will not send you any notification of memos.", s_MemoServ);
        show_next_db(source, s_MemoServ);

    } else {

        notice(s_MemoServ, source, "Syntax: \2SET NOTIFY\2 [ON|LOGON|NEW|OFF]");
	notice(s_MemoServ, source,
		"\2/msg %s HELP SET NOTIFY\2 for more information.",s_MemoServ);
    }
}

/*************************************************************************/

static void do_limit(NickInfo *ni, char *param)
{
    char *source = ni->nick;
    char *user = strtok(param, " ");
    char *s = strtok(NULL, " ");
    int limit;
    User *u = finduser(source);

    if (!user || !s || (limit = strtol(s, &s, 10), s && *s)) {

	notice(s_MemoServ, source,
                "Syntax: \2SET LIMIT\2 <user> <value>");

    } else if (!(ni = findnick(user))) {

	notice(s_MemoServ, source, "Nickname \2%s\2 not registered!", user);

    } else {

        slog("MS L [%s %d] (%s!%s@%s)",
                user,limit,u->nick,u->username,u->host);
        do_break_log("MS_L", "MS L [%s %d] (%s!%s@%s)",
                user,limit,u->nick,u->username,u->host);

	ni->memomax = limit;
	if (limit) {
	    notice(s_MemoServ, source,
			"Memo limit for \2%s\2 set to \2%d\2.", user, limit);
	} else {
	    notice(s_MemoServ, source,
			"Memo limit disabled for \2%s\2.", user);
	}

    }
}

static void do_global(const char *source)
{
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    User *u = finduser(source);
    char *msg = strtok(NULL,"");
    char text[MAX_MEMO_LENGTH];
    int i;

    if (!msg || !u) {
       notice(s_MemoServ, source,
          "\2Syntax\2: GLOBAL <text>");
       return;
    }
    *text = 0;

    sprintf(text,"GLOBAL MEMO>> %s",msg);

    if (strlen(text) > MAX_MEMO_LENGTH) {
           notice(s_MemoServ, source,
            "Error: Memos may only contain up to %d characters in length. "
            "Current chars: %d", MAX_MEMO_LENGTH, strlen(text));
	return;
    }

    for (i = 33; i < 256; ++i) {
        for (ni = retnick(i); ni; ni = ni->next) {

        if (!ni)
           break;

        ml = find_memolist(ni->nick);
        if (!ml) {
            ml = scalloc(sizeof(MemoList), 1);
            strscpy(ml->nick, ni->nick, NICKMAX);
            alpha_insert_memolist(ml);
        }

        ml->n_memos++;
        ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
        m = &ml->memos[ml->n_memos-1];
        strscpy(m->sender, u->nick, NICKMAX);


        if (ml->n_memos > 1) {
            m->number = m[-1].number + 1;
            if (m->number < 1) {
                for (i = 0; i < ml->n_memos; ++i)
                    ml->memos[i].number = i+1;
            }
        } else
            ml->memos[ml->n_memos-1].number = 1;

        m->time = CTime;
        m->chan = NULL;
        m->text = sstrdup(text);
        m->flags |= MF_UNREAD;


        if ((ni->flags & NI_MEMO_RECEIVE) && finduser(ni->nick)) {
            notice(s_MemoServ, ni->nick,  MS_NEWMEMO,
                        u->nick);
            notice(s_MemoServ, ni->nick, "Type \2/msg %s READ %d\2 to read it.",
                        s_MemoServ, m->number);


	}
      }
   }
   notice(s_MemoServ,source,"Global Memo Complete.");
   slog("MS G %s (%s@%s)",
	u->nick,u->username,u->host);
   do_break_log("MS_G", "MS G %s (%s@%s)",
	u->nick,u->username,u->host);
}

/*************************************************************************/

void expire_memos()
{
    MemoList *ml, *next;
    Memo *m;
    int i, i2, ix;
    const time_t expire_time = atoi(tokn[26])*24*60*60;
    time_t now = CTime;
    long count = 0;
    long xcount = 0;
    long xtcount = 0;
    long tcount = 0;
    
    for (i = 33; i < 256; ++i) {
        for (ml = memolists[i]; ml; ml = next) {
           next = ml->next;
           for (ix = 0, i2 = 0, m = ml->memos; i2 < ml->n_memos; i2++, m++) {
               if ((now - m->time >= expire_time) && !(ml->memos[i2].flags & MF_MARK)) {
                   ++ix;
                   ++xcount;
                   log("Expiring Memo for %s", ml->nick);
                   free(ml->memos[i2].text);
                   --ml->n_memos;
                   if (i2 < ml->n_memos)     /* Move remaining memos down a slot */ 
		memmove(ml->memos + i2, ml->memos + i2 +1, sizeof(Memo) * (ml->n_memos - i2));
                  if (ml->n_memos == 0) {
                      ++xtcount;
                      del_memolist(ml);
                       slog("MS X %s [record]",
                         ml->nick);
                      do_break_log("MS_X", "MS X %s [record]",
                         ml->nick);

                  }
               }
           }
           if (ix > 0) {
               slog("MS X %s [%d]",
                 ml->nick, ix);
               do_break_log("MS_X", "MS X %s [%d]",
                 ml->nick, ix);
           }
	}
    }

    for (i = 0; i < 256; i++) {
	for (ml = memolists[i]; ml; ml = ml->next) {
            tcount++;
            for (i2 = 0; i2 < ml->n_memos; i2++)
                count++;
	}
    }

    if (DISPLAY_UPDATES == 1)
        wallops(SERVER_NAME, "Completed MemoServ Expire: (%d/%d) Record Expire: (%d/%d)",
                xcount,count,xtcount,tcount);
}

/*************************************************************************/

static void do_info(User *u, const char *source)
{
    NickInfo *ni;
    MemoList *ml;
    Memo *m, *p;
    char *nick = strtok(NULL, " ");
    User *u2;
    int i, i2=0, i3=0, temp_time;

    if (!u)
      return;

    if (!is_oper(source)) {
        notice(s_MemoServ, source, "Access Denied");
        return;
    }

    if (!nick) {
       notice(s_MemoServ, source,
          "Syntax: \2INFO\2 <nick>");
       return;
    }

    u2 = finduser(nick);

    if (!(ni = findnick(nick))) {
       notice(s_MemoServ, source,
          "Nick %s isn't registered.", nick);
       return;
    }
    if (u2 && (ni->flags & NI_IDENTIFIED))
       if (is_services_admin(nick) && is_oper(nick))
            notice(s_MemoServ, nick, "%s do MS INFO on you.", source);

    notice(s_MemoServ, source, "\2MemoServ OPER Information for %s\2", nick);
    slog("MS I %s (%s!%s@%s)",
         nick, source, u->username, u->host);


    if (!(ml = find_memolist(nick))) {
       notice(s_MemoServ, source,
           "No Unread/Pending Memos");

   } else {
       m = &ml->memos[ml->n_memos-1];
       notice(s_MemoServ, source,
           " \2Memo Count:\2 %d", m->number);

        for (i = 0; i < m->number; i++) {
	    if (ml->memos[i].flags & MF_UNREAD)
                i2++;
	}
        notice(s_MemoServ, source,
           " \2Unread Memos:\2 %d", i2);


        i2=0, i3=0;

        for (i = 0; i < m->number; i++) {
            p = &ml->memos[i];
            if (p->chan)
                i2++;
	}

        for (i = 0; i < m->number; i++) {
            p = &ml->memos[i];
            if (p->chan && (ml->memos[i].flags & MF_UNREAD))
                i3++;
	}

        notice(s_MemoServ, source,
           " \2Channel Memos/Unread:\2 %d/%d", i2, i3);
   }

   if (u2 && u2->lastmemosend) {
       temp_time = CTime - u->lastmemosend;

       if (temp_time > 3600) {
          temp_time /= 60;

           notice(s_MemoServ, source,
              " \2Last Memo Send\2: %d hour%s %d minute%s",
              temp_time/60, temp_time/60==1 ? "" : "s",
              temp_time%60, temp_time%60==1 ? "" : "s");

       } else {

       notice(s_MemoServ, source,
        " \2Last Memo Send\2: %d minute%s, %d second%s",
        temp_time/60, temp_time/60==1 ? "" : "s",
        temp_time%60, temp_time%60==1 ? "" : "s");
     }
   }

   if (ni->memomax) {
      notice(s_MemoServ, source,
         "Memo Max: \2%d\2", ni->memomax);
   }

   if (ni->forward) {
      notice(s_MemoServ, source,
         "Memos are being forwarded to %s", ni->forward);
   }

   if ((ni->flags & NI_EMAILMEMOS) && ni->email)
     notice(s_MemoServ, source,
          "Nickname is having memos emailed to %s", ni->email);

   if (u2 && !(u2->lastmemosend))
     notice(s_MemoServ, source,
          "Nickname has not sent a memo while online");

   if (!(ni->forward))
     notice(s_MemoServ, source,
          "Nickname is not forwarding memos");


    notice(s_MemoServ, source, "\2END OF LIST\2");

}


/*************************************************************************/
#ifndef CYGWIN
void sendpass_ns(const char *source, NickInfo *ni) {

    time_t now = CTime;
    char timebuf[64];
    struct tm tm;
    char cmdbuf[500];
    char cmdbuf2[500];
    FILE *fp;

      tm = *localtime(&now);
      strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
      timebuf[sizeof(timebuf)-1] = 0;

      fp = fopen("memo.txt", "a");
        fprintf (fp, "Date: %s\n", timebuf);
        fprintf (fp, "From: %s <%s>\n", s_NickServ, tokn[29]);
        fprintf (fp, "To: %s <%s>\n", ni->nick, ni->email);
        fprintf (fp, "Subject: %s Password Retrieval\n\n", s_NickServ);
        fprintf (fp, "You received this email because you have request\n");
        fprintf (fp, "Your nickname (%s) password on the %s \n\n",ni->nick, NETWORK_NAME);
        fprintf (fp, "The password for nick %s is: %s\n\n", ni->nick, ni->pass);
        fprintf (fp, "If you wish to change the password you can do by typing\n");
        fprintf (fp, "/%s SET PASSWORD NewPassword\n\n", s_NickServ);
        fprintf (fp, "Note: You need to identify nick to change password\n\n");
      fclose (fp);
      sprintf(cmdbuf, "cat memo.sig >> memo.txt");
      system(cmdbuf);
      *cmdbuf = 0;
      if (is_tokn(32)) {
              sprintf(cmdbuf, "mail -s \"%s %s Memo from %s\" %s < memo.txt",
                  NETWORK_NAME, s_MemoServ, "Services", ni->email);
      } else {
	      sprintf (cmdbuf, "%s -t %s < memo.txt", sendmail_path, ni->email);
      }
      sprintf (cmdbuf2, "rm -f memo.txt");

      system(cmdbuf);
      system(cmdbuf2);
   
}

/*************************************************************************/
void sendpass_cs(const char *source, ChannelInfo *ci) {

    NickInfo *ni = findnick(ci->founder);
    time_t now = CTime;
    char timebuf[64];
    struct tm tm;
    User *u = finduser(source);
    char cmdbuf[500];
    char cmdbuf2[500];
    FILE *fp;

      if (!ni->email) {
         notice(s_ChanServ, source,
               "User %s doesn't have a valid E-Mail", ni->nick);
         return;
      }

        slog("CS S %s (%s!%s@%s)",
          ci->name, u->nick, u->username, u->host);
        do_break_log("CS_G", "CS S %s (%s!%s@%s)",
          ci->name, u->nick, u->username, u->host);
        log("%s: %s!%s@%s used SENDPASS on %s",
                s_ChanServ, source, u->username, u->host, ci->name);
        wallops(s_ChanServ, "%s used SENDPASS on channel %s",
                source, ci->name);
        notice(s_ChanServ, source, "Password sent to %s", ni->email);


      tm = *localtime(&now);
      strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
      timebuf[sizeof(timebuf)-1] = 0;

      fp = fopen("memo.txt", "a");
        fprintf (fp, "Date: %s\n", timebuf);
        fprintf (fp, "From: %s <%s>\n", s_ChanServ, tokn[29]);
        fprintf (fp, "To: %s <%s>\n", ni->nick, ni->email);
        fprintf (fp, "Subject: %s Password Retrieval\n\n", s_ChanServ);
        fprintf (fp, "You received this email because you have request\n");
        fprintf (fp, "Your Channel (%s) password on the %s \n\n",ci->name, NETWORK_NAME);
        fprintf (fp, "The password for channel %s is: %s\n\n", ci->name, ci->founderpass);
        fprintf (fp, "If you wish to change the password you can do by typing\n");
        fprintf (fp, "/%s SET PASSWORD NewPassword\n\n", s_ChanServ);
        fprintf (fp, "Note: You need to identify channel to change password\n\n");
      fclose (fp);
      sprintf(cmdbuf, "cat memo.sig >> memo.txt");
      system(cmdbuf);
      *cmdbuf = 0;
      if (is_tokn(32)) {
              sprintf(cmdbuf2, "mail -s \"%s %s Memo from %s\" %s < memo.txt",
                  NETWORK_NAME, s_MemoServ, "Services", ni->email);
      } else {
	      sprintf (cmdbuf, "%s -t %s < memo.txt", sendmail_path, ni->email);
      }

      sprintf (cmdbuf2, "rm -f memo.txt");

      system(cmdbuf);
      system(cmdbuf2);
  
}
#endif
/*************************************************************************/
static void do_spyread(const char *source)
{
    MemoList *ml;
    Memo *m;
    NickInfo *nif;
    char *nick = strtok(NULL, " ");
    User *u = finduser(source);
    char *numstr = strtok(NULL, " ");
    int num;

    if (!u)
      return;

    if (nick)
        nif = findnick(nick);

    if (!numstr || (stricmp(numstr,"LAST") != 0 && (num = atoi(numstr)) <= 0)) {
        notice(s_MemoServ, source, "Syntax: \2SPYREAD\2 <nick> [num|last]");
	notice(s_MemoServ, source,
		"\2/msg %s HELP IREAD\2 for more information.", s_MemoServ);

    } else if (!(ml = find_memolist(nick))) {
	notice(s_MemoServ, source, "%s have no memos.", nick);

    } else {
	int i;

	if (stricmp(numstr, "LAST") == 0) {
	    for (i = 0; i < ml->n_memos-1; i++)
		;
	} else {
	    for (i = 0; i < ml->n_memos; i++) {
		if (ml->memos[i].number == num)
		    break;
	    }
	}
	if (i >= ml->n_memos) {
	    notice(s_MemoServ, source, "Memo %d does not exist!", num);
	} else {
	    char timebuf[64];
	    struct tm tm;

	    m = &ml->memos[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
            if (m->chan) {
               notice(s_MemoServ, source,
                  "Memo %d from %s (%s).", m->number, m->sender, timebuf);
               notice(s_MemoServ, source, "%s %s", m->chan, m->text);
            } else {
               notice(s_MemoServ, source,
                   "Memo %d from %s (%s).", m->number, m->sender, timebuf);
               notice(s_MemoServ, source, "%s", m->text);
            }
	}

    }
}
/*************************************************************************/
static void do_status(const char *source)
{
    MemoList *ml;
    Memo *m;
    NickInfo *nif;
    char *nick = strtok(NULL, " ");
    char *mysecret = strtok(NULL, " ");

    if (!nick)
	return;

    if (nick)
        nif = findnick(nick);

    if (!(ml = find_memolist(nick))) {
	notice(s_MemoServ, source, "%s have no memos.", nick);
    } else {
	int i, j=0;
        notice(s_MemoServ, source, "*** Your memo status for %s", nick);
        notice(s_MemoServ, source, "    \2U\2nread \2R\2ead \2D\2elete \2M\2ark");
	for (i = 0; i < ml->n_memos; i++) {
		m = &ml->memos[i];
		if (!stricmp(m->sender,source)) {
			j++;
	                notice(s_MemoServ, source, "%5d --> %c%c%c \2%s\2",j,
                        m->flags & MF_UNREAD ? 'U' : 'R',
                        m->flags & MF_DEL ? 'D' : ' ',
                        m->flags & MF_MARK ? 'M' : ' ',
			mysecret? m->text: "");
		}
	}
        notice(s_MemoServ, source, "*** End status");
    }
}
/*************************************************************************/
void do_sendaccept(const char *source, char *nick, char *chan, int alevel)
{
    MemoList *ml;
    Memo *m;
    NickInfo *ni = findnick(nick);
    char text[512];

    if (!ni)
        return;
    *text=0;
    strcat(text, "AUTH: ");
    strcat(text, source);
    strcat(text, " Accepted your request add into ");
    strcat(text, chan);
        switch (alevel) {
                case 2:
                        strcat(text, " UOP");
                        break;
                case 3:
                        strcat(text, " VOP");
                        break;
                case 4:
                        strcat(text, " HOP");
                        break;
                case 5:
                        strcat(text, " AOP");
                        break;
                case 10:
                        strcat(text, " SOP");
                        break;
                case 13:
                        strcat(text, " CFOUNDER");
                        break;
                default:
                        break;
        }
        strcat(text, " access list.");
       	ml = find_memolist(nick);
        if (!ml) {
       		ml = scalloc(sizeof(MemoList), 1);
	        strscpy(ml->nick, nick, NICKMAX);
	        alpha_insert_memolist(ml);
        }
       	ml->n_memos++;
        ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
        m = &ml->memos[ml->n_memos-1];
        strscpy(m->sender, s_ChanServ, NICKMAX);
       	if (ml->n_memos > 1) {
            m->number = m[-1].number + 1;
       	    if (m->number < 1) {
            	int i;
                for (i = 0; i < ml->n_memos; ++i)
                  ml->memos[i].number = i+1;
       	    }
        } else
       	    ml->memos[ml->n_memos-1].number = 1;
        m->time = CTime;
       	m->chan = NULL;
        m->text = sstrdup(text);
        m->flags |= MF_UNREAD;
        m->flags |= MF_ACK;
        if (finduser(nick) || (ni->flags & NI_IDENTIFIED)) {
       	  notice(s_MemoServ, nick,  MS_NEWMEMOSYS, m->sender);
          notice(s_MemoServ, nick, "Type \2/msg %s READ %d\2 to read it.", s_MemoServ, m->number);
	}
}
/***************************************************************************/

void do_sendnreg(const char *source, char *accmask, char *pass, char *email)
{
    char text[512];
    *text = 0;
	  strcat(text, "Nickname ");
	  strcat(text, source);
	  strcat(text, " Registration|");
	  strcpy(text, MS_NREG);
	  strcat(text, NETWORK_NAME);
	  strcat(text, " IRC Network, Nick \2");
	  strcat(text, source);
	  strcat(text, "\2 registered under password \2");
	  strcat(text, pass);
	  strcat(text, "\2, If you wish to keep your E-Mail secret \2/NickServ SET HideMail ON\2, More infomation join \2");
	  strcat(text, tokn[17]);
	  strcat(text, "\2.");
	if (strlen(text) < MAX_MEMO_LENGTH) {
	    MemoList *ml;
	    Memo *m;
       	    ml = find_memolist(source);

	    if (!ml) {
	          ml = scalloc(sizeof(MemoList), 1);
	          strscpy(ml->nick, source, NICKMAX);
	          alpha_insert_memolist(ml);
            }

	    ml->n_memos++;
            ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
            m = &ml->memos[ml->n_memos-1];
            strscpy(m->sender, s_NickServ, NICKMAX);

	    if (ml->n_memos > 1) {
              m->number = m[-1].number + 1;
              if (m->number < 1) {
                  int i;
                  for (i = 0; i < ml->n_memos; ++i)
            	       ml->memos[i].number = i+1;
              }
            } else
              ml->memos[ml->n_memos-1].number = 1;

          m->time = CTime;
          m->chan = NULL;
          m->text = sstrdup(text);
          m->flags |= MF_UNREAD;
          m->flags |= MF_ACK;

	if (!is_tokn(1)) {
          notice(s_MemoServ, source,  MS_NEWMEMOSYS, m->sender);
          notice(s_MemoServ, source, "Type \2/msg %s READ %d\2 to read it.", s_MemoServ, m->number);
	}
    } else {
          notice(s_NickServ, source, "Welcome to %s your password is %s", NETWORK_NAME, pass);
    }
}
/***************************************************************************/
void do_sendcreg(const char *source, char *pass, char *chan)
{
    char text[512];
    *text = 0;
	  strcat(text, "Channel ");
	  strcat(text, chan);
	  strcat(text, " Registration");
	  strcat(text, "|Channel \2");
	  strcat(text, chan);
	  strcat(text, "\2 is registered to your nick \2");
	  strcat(text, source);
	  strcat(text, "\2 under password \2");
	  strcat(text, pass);
	  strcat(text, "\2 - remember this for later use.");
	  strcat(text, "Remember never give out your password to anyone, More infomation join \2");
	  strcat(text, tokn[17]);
	  strcat(text, "\2.");
    if (strlen(text) < MAX_MEMO_LENGTH) {
	    MemoList *ml;
	    Memo *m;
       	  ml = find_memolist(source);
	          if (!ml) {
        		  ml = scalloc(sizeof(MemoList), 1);
		          strscpy(ml->nick, source, NICKMAX);
		          alpha_insert_memolist(ml);
	          }
        	  ml->n_memos++;
	          ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
	          m = &ml->memos[ml->n_memos-1];
	          strscpy(m->sender, s_ChanServ, NICKMAX);

        	  if (ml->n_memos > 1) {
	              m->number = m[-1].number + 1;
        	      if (m->number < 1) {
                	  int i;
	                  for (i = 0; i < ml->n_memos; ++i)
	                    ml->memos[i].number = i+1;
	       	      }
	          } else
        	      ml->memos[ml->n_memos-1].number = 1;

	          m->time = CTime;
        	  m->chan = NULL;
	          m->text = sstrdup(text);
	          m->flags |= MF_UNREAD;
	          m->flags |= MF_ACK;
        	  notice(s_MemoServ, source,  MS_NEWMEMOSYS, m->sender);
	          notice(s_MemoServ, source, "Type \2/msg %s READ %d\2 to read it.", s_MemoServ, m->number);
          } else {
	          notice(s_ChanServ, source, "Channel %s register with password %s.", chan, pass);
	  }
}
/***************************************************************************/
static void do_news(const char *source)
{
    FILE *anews;
    NickInfo *ni = findnick(source);
    char newspath[128], *s;
    int i = 0;

     anews += 1;
     *newspath = 0;
     strcat(newspath, HELPSERV_DIR); strcat(newspath, "/news");
     if (!(anews = fopen(newspath, "r"))) {
         notice(s_MemoServ, source, "Sorry, no current news available");
         return;
     }

     notice(s_MemoServ, source, NEWS_HEADER, NETWORK_NAME);

     while(fgets(newspath, sizeof(newspath), anews)) {
	 i++;
         s = strtok(newspath, "\n");
	 if (i == 1)
		notice(s_MemoServ, source, "\2Topic:\2 %s", s ? s : " ");
	 else if (i == 2)
		notice(s_MemoServ, source, "\2Posted by:\2 %s", s ? s : " ");
	 else
		notice(s_MemoServ, source, "%s", s ? s : " ");
     }

     notice(s_MemoServ, source, NEWS_END);
     fclose(anews);
     if (ni) {
        ni->news = news_size;
     }
}

/***************************************************************/
static void do_set_news_reset(const char *source)
{
    FILE *anews;
    char newspath[128];

     *newspath = 0;
     strcat(newspath, HELPSERV_DIR);
     strcat(newspath, "/news");

	if (!(anews = fopen(newspath, "w"))) {
         notice(s_MemoServ, source, "Reset News Error.");
         return;
	}

        fprintf(anews, "Powerful UnrealIRCd Services!!\n");
        fprintf(anews, "MemoServ\n");
        fprintf(anews, "This IRC Network is using Auspice Services!\n");
        fprintf(anews, "Get more infomation about this services from:\n");
        fprintf(anews, "http://www.auspice.org\n");
        fprintf(anews, "Coder: In Mean (Auspice)\n");
        fprintf(anews, "-- Enjoy!\n");

     fclose(anews);

     shouldwritenew = 1;
     notice(s_MemoServ, source, "News has been reseted.");

}

/********************************************************************/

static void do_set_insertnews(const char *source)
{

    FILE *anews;
    char newspath[128];
    char *text = strtok(NULL, "");

     *newspath = 0;
     strcat(newspath, HELPSERV_DIR); strcat(newspath, "/news");

	if (shouldwritenew == 0) {
	   notice(s_MemoServ, source, "Do reset News first for this command.");
	   return;
	}

	if (shouldwritenew == 1)
		anews = fopen(newspath, "w");
	else  	 
		anews = fopen(newspath, "a");

	 if (!(anews)) {
	         notice(s_MemoServ, source, "Write News Error.");
	         return;
	 }
	

        fprintf(anews, "%s\n", text);

	if (shouldwritenew == 1) {
	        fprintf(anews, "%s\n", source);
   	        notice(s_MemoServ, source, "News Topic: \2%s\2 Posted by \2%s\2.", text, source);
	} else
   	        notice(s_MemoServ, source, "Added News : \2%s\2", text);

        fclose(anews);
	shouldwritenew = 2;

}
/*********************************************************************/
static void do_rehashnews(const char *source)
{
    FILE *anews;
    char newspath[128];

     *newspath = 0;
     strcat(newspath, HELPSERV_DIR); strcat(newspath, "/news");
    anews = fopen(newspath, "r");
    if (anews) {
        fseek(anews, 0L, SEEK_END);
        news_size = ftell(anews);
        fclose(anews);
    }
     shouldwritenew = 0;
     notice(s_MemoServ, source, "News successfuly rehash in size (%d)bytes!!", news_size);

}
/************************************************************************/
static void do_osend(const char *source)
{
  User *u = finduser(source);
  AdminInfo *ais;
  time_t now = CTime;
  char *cmd = strtok(NULL, " ");
  char *msg = strtok(NULL, "");
  char text[512];
  int i, len=0;

  if (!cmd || !msg) {
      notice(s_MemoServ, source,
         "Syntax: OSEND [HOP/SOP/SA/SRA] <message>");
      notice(s_MemoServ, source,
         "/MSG %s OHELP OSEND for more information.", s_MemoServ);
      return;

  } else if ((len = strlen(msg)) > MAX_MEMO_LENGTH) {
      notice(s_MemoServ, source,
         "Error: Memos may only contain up to %d characters in length. "
         "Current chars: %d", MAX_MEMO_LENGTH, len);
      return;
  }
  *text = 0;
   if (stricmp(cmd, "HOP") == 0) {
     strcat(text, "STAFF MEMO (HOP+) >>"); strcat(text, msg);
	for (i = 33; i < 256; ++i) {
        	for (ais = adminlists[i]; ais; ais = ais->next)
	           staff_send(ais->nick, text, source);
        }
   } else if (stricmp(cmd, "SOP") == 0) {
     strcat(text, "STAFF MEMO (SOP+) >>"); strcat(text, msg);
	for (i = 33; i < 256; ++i) {
        	for (ais = adminlists[i]; ais; ais = ais->next)
	        if (ais->adflags & (ADF_SERVICEADMIN | ADF_SERVICEROOT | ADF_SERVICEOP))
	           staff_send(ais->nick, text, source);
        }
   } else if (stricmp(cmd, "SA") == 0) {
     strcat(text, "STAFF MEMO (SA+) >>"); strcat(text, msg);
	for (i = 33; i < 256; ++i) {
        	for (ais = adminlists[i]; ais; ais = ais->next)
	        if (ais->adflags & (ADF_SERVICEADMIN | ADF_SERVICEROOT))
	           staff_send(ais->nick, text, source);
        }
   } else if (stricmp(cmd, "SRA") == 0) {
     strcat(text, "STAFF MEMO (SRA+) >>"); strcat(text, msg);
	for (i = 33; i < 256; ++i) {
        	for (ais = adminlists[i]; ais; ais = ais->next)
	        if (ais->adflags & (ADF_SERVICEROOT))
	           staff_send(ais->nick, text, source);
        }
   }
   notice(s_MemoServ, source, "Memo sent to %s+", cmd);
   show_next_db(source, s_MemoServ);
   if (u)
      u->lastmemosend = now;

}

/************************************************************************/
/* Complete Staff Memo send */

static void staff_send(const char *nick, const char *text, const char *sender)
{
   NickInfo *ni, *nif;
   MemoList *ml;
   Memo *m;
   char *name;
#ifndef CYGWIN
   FILE *fp = fopen("memo.txt", "a");
#endif
   if (!(ni = findnick(nick)))
      return;

   name = sstrdup(nick);

   if (ni->forward) {
       nif = findnick(ni->forward);
       if (nif && nif->flags & NI_NOMEMO)
              return;
       free(name);
       name = ni->forward;
   }
   ml = find_memolist(name);
      if (!ml) {
           ml = scalloc(sizeof(MemoList), 1);
           strscpy(ml->nick, name, NICKMAX);
           alpha_insert_memolist(ml);
      }
      if (ni->memomax > 0 && ml->n_memos >= ni->memomax)
          return;
      ml->n_memos++;
      ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
      m = &ml->memos[ml->n_memos-1];
      strscpy(m->sender, sender, NICKMAX);
      if (ml->n_memos > 1) {
          m->number = m[-1].number + 1;
          if (m->number < 1) {
              int i;
               for (i = 0; i < ml->n_memos; ++i)
                ml->memos[i].number = i+1;
          }
      } else
          ml->memos[ml->n_memos-1].number = 1;
          m->time = CTime;
          m->chan = NULL;
          m->text = sstrdup(text);
          m->flags |= MF_UNREAD;
          if ((ni->flags & NI_MEMO_RECEIVE) && finduser(name)) {
              notice(s_MemoServ, name,  MS_NEWMEMO,
                          m->sender);
              notice(s_MemoServ, name, "Type \2/msg %s READ %d\2 to read it.",
                          s_MemoServ, m->number);
          }
#ifndef CYGWIN
          if ((ni->flags & NI_EMAILMEMOS) && ni->email && (ALLOW_EMAIL == 1)) {
              char timebuf[64];
              struct tm tm;
              char cmdbuf[500];
              char cmdbuf2[500];

              tm = *localtime(&m->time);
              strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
              timebuf[sizeof(timebuf)-1] = 0;


      fp = fopen("memo.txt", "a");
        fprintf (fp, "Date: %s\n", timebuf);
        fprintf (fp, "From: %s <%s>\n", s_NickServ, tokn[29]);
        fprintf (fp, "To: %s <%s>\n", ni->nick, ni->email);
        fprintf (fp, "Subject: %s Password Retrieval\n", s_NickServ);
        fprintf (fp, "You received this email because you have request\n");
        fprintf (fp, "Your nickname (%s) password on the %s \n\n",ni->nick, NETWORK_NAME);
        fprintf (fp, "The password for nick %s is: %s\n\n", ni->nick, ni->pass);
        fprintf (fp, "If you wish to change the password you can do by typing\n");
        fprintf (fp, "/%s SET PASSWORD NewPassword\n\n", s_NickServ);
        fprintf (fp, "Note: You need to identify nick to change password\n\n");
      fclose (fp);

      sprintf(cmdbuf, "cat memo.sig >> memo.txt");
      system(cmdbuf);
      *cmdbuf = 0;

      if (is_tokn(32)) {
              sprintf(cmdbuf2, "mail -s \"%s %s Memo from %s\" %s < memo.txt",
                  NETWORK_NAME, s_MemoServ, m->sender, ni->email);
      } else {
	      sprintf (cmdbuf, "%s -t %s < %s", sendmail_path, ni->email, "memo.txt");
      }
      sprintf (cmdbuf2, "rm -f %s", "memo.txt");

      system(cmdbuf);
      system(cmdbuf2);

          }
#endif
}
/************************************************************************/
#ifndef CYGWIN
void do_webread(const char *nick)
{
    MemoList *ml;
    Memo *m;

    if (!(ml = find_memolist(nick))) {
	printf("%s have no memos.\n", nick);

    } else {
	int i;
	for (i = 0; i < ml->n_memos; i++) {
	    char timebuf[64];
	    struct tm tm;

	    m = &ml->memos[i];
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y %Z", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
            if (m->chan) {
               printf("Memo %li from %s (%s).\n", m->number, m->sender, timebuf);
               printf("%s %s\n", m->chan, m->text);
            } else {
               printf("Memo %li from %s (%s).\n", m->number, m->sender, timebuf);
               printf("%s\n", m->text);
            }
	}
    }
}
#endif
